@@ -49,13 +49,14 @@ import Network.TLS.Handshake.State
 import Network.TLS.Handshake.State13
 import Network.TLS.PostHandshake
 import Network.TLS.KeySchedule
-import Network.TLS.Types (Role(..), HostName)
+import Network.TLS.Types (Role(..), HostName, AnyTrafficSecret(..), ApplicationSecret)
 import Network.TLS.Util (catchException, mapChunks_)
 import Network.TLS.Extension
 import qualified Network.TLS.State as S
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as C8
 import qualified Data.ByteString.Lazy as L
+import           Control.Monad (unless, when)
 import qualified Control.Exception as E
 
 import Control.Monad.State.Strict
@@ -101,7 +102,8 @@ sendData ctx dataToSend = liftIO $ do
         -- All chunks are protected with the same write lock because we don't
         -- want to interleave writes from other threads in the middle of our
         -- possibly large write.
-        mapM_ (mapChunks_ 16384 sendP) (L.toChunks dataToSend)
+        let len = ctxFragmentSize ctx
+        mapM_ (mapChunks_ len sendP) (L.toChunks dataToSend)
 
 -- | Get data out of Data packet, and automatically renegotiate if a Handshake
 -- ClientHello is received.  An empty result means EOF.
@@ -174,7 +176,13 @@ recvData13 ctx = do
                     terminate (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
               Established         -> return x
               NotEstablished      -> throwCore $ Error_Protocol ("data at not-established", True, UnexpectedMessage)
-        process ChangeCipherSpec13 = recvData13 ctx
+        process ChangeCipherSpec13 = do
+            established <- ctxEstablished ctx
+            if established /= Established then
+                recvData13 ctx
+              else do
+                let reason = "CSS after Finished"
+                terminate (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
         process p             = let reason = "unexpected message " ++ show p in
                                 terminate (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
 
@@ -194,7 +202,7 @@ recvData13 ctx = do
             -- session manager).
             withWriteLock ctx $ do
                 Just resumptionMasterSecret <- usingHState ctx getTLS13ResumptionSecret
-                (_, usedCipher, _) <- getTxState ctx
+                (_, usedCipher, _, _) <- getTxState ctx
                 let choice = makeCipherChoice TLS13 usedCipher
                     psk = derivePSK choice resumptionMasterSecret nonce
                     maxSize = case extensionLookup extensionID_EarlyData exts >>= extensionDecode MsgTNewSessionTicket of
@@ -208,6 +216,9 @@ recvData13 ctx = do
                 -- putStrLn $ "NewSessionTicket received: lifetime = " ++ show life ++ " sec"
             loopHandshake13 hs
         loopHandshake13 (KeyUpdate13 mode:hs) = do
+            when (ctxQUICMode ctx) $ do
+                let reason = "KeyUpdate is not allowed for QUIC"
+                terminate (Error_Misc reason) AlertLevel_Fatal UnexpectedMessage reason
             checkAlignment hs
             established <- ctxEstablished ctx
             -- Though RFC 8446 Sec 4.6.3 does not clearly says,
@@ -267,10 +278,8 @@ onError :: Monad m => (TLSError -> AlertLevel -> AlertDescription -> String -> m
                    -> TLSError -> m B.ByteString
 onError _ Error_EOF = -- Not really an error.
             return B.empty
-onError terminate err@(Error_Protocol (reason,fatal,desc)) =
-    terminate err (if fatal then AlertLevel_Fatal else AlertLevel_Warning) desc reason
-onError terminate err =
-    terminate err AlertLevel_Fatal InternalError (show err)
+onError terminate err = let (lvl,ad) = errorToAlert err
+                        in terminate err lvl ad (errorToAlertMessage err)
 
 terminateWithWriteLock :: Context -> ([(AlertLevel, AlertDescription)] -> IO ())
                        -> TLSError -> AlertLevel -> AlertDescription -> String -> IO a
@@ -293,13 +302,15 @@ recvData' :: MonadIO m => Context -> m L.ByteString
 recvData' ctx = L.fromChunks . (:[]) <$> recvData ctx
 
 keyUpdate :: Context
-          -> (Context -> IO (Hash,Cipher,C8.ByteString))
-          -> (Context -> Hash -> Cipher -> C8.ByteString -> IO ())
+          -> (Context -> IO (Hash,Cipher,CryptLevel,C8.ByteString))
+          -> (Context -> Hash -> Cipher -> AnyTrafficSecret ApplicationSecret -> IO ())
           -> IO ()
 keyUpdate ctx getState setState = do
-    (usedHash, usedCipher, applicationSecretN) <- getState ctx
+    (usedHash, usedCipher, level, applicationSecretN) <- getState ctx
+    unless (level == CryptApplicationSecret) $
+        throwCore $ Error_Protocol ("tried key update without application traffic secret", True, InternalError)
     let applicationSecretN1 = hkdfExpandLabel usedHash applicationSecretN "traffic upd" "" $ hashDigestSize usedHash
-    setState ctx usedHash usedCipher applicationSecretN1
+    setState ctx usedHash usedCipher (AnyTrafficSecret applicationSecretN1)
 
 -- | How to update keys in TLS 1.3
 data KeyUpdateRequest = OneWay -- ^ Unidirectional key update
